from __future__ import annotations

import urllib.parse
import tooltime


def parse_url(
    url_search: str,
    display: dict[str, str],
    defaults: dict[str, str],
) -> None:
    parsed = urllib.parse.parse_qs(urllib.parse.urlparse(url_search).query)
    for key in defaults.keys():
        if key in parsed:
            display[key] = parsed[key][0]
    if 'now' in parsed:
        display['now'] = parsed['now'][0]


def create_new_url(display: dict[str, str], defaults: dict[str, str]) -> str:
    params: list[str] = []
    for key, value in defaults.items():
        if display[key] != value:
            params.append(key + '=' + display[key])

    current_now = display.get('now')
    if current_now is not None and tooltime.timestamp_to_seconds(
        current_now[:10]
    ) != tooltime.timestamp_to_seconds(defaults['now']):
        params.append('now=' + tooltime.timestamp_to_date(current_now[:10]))

    return '?' + '&'.join(params) if params else ''



from __future__ import annotations

import plotly.graph_objects as go  # type: ignore
from dash import Dash, Input, Output, State  # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from . import dashboard
from . import inputs as inputs_module
from . import ui_types
from . import urls
from . import state as state_module


def _create_app(
    name: str,
    dashboard: dashboard.Dashboard,
    assets_folder: str | None = None,
    debug: bool = False,
    external_stylesheets: list[str] = [dbc.themes.BOOTSTRAP],
) -> Dash:
    # create app
    if assets_folder is None:
        assets_folder = '/home/storm/repos/stables/assets'
    app = Dash(
        name,
        assets_folder=assets_folder,
        external_stylesheets=external_stylesheets,
    )
    app.title = name
    n_keydowns = 0

    # create inputs
    inputs = inputs_module.create_inputs(dashboard.spec, dashboard.dfs)

    # create layout
    app.layout = dashboard.get_layout(inputs)

    # add serverside callbacks
    update_ui_args, update_chart_args = _compute_decorator_args(dashboard.spec)

    @app.callback(*update_ui_args)  # type: ignore
    def update_ui(*args) -> list[str | bool]:
        # parse input args
        *raw_display, url, initial_load, _n_keydowns, shortcut, help_open = args
        display = dict(zip(list(inputs.keys()), raw_display))

        # load from url
        if initial_load:
            urls.parse_url(url, display, dashboard.spec['default_state'])
            initial_load = False

        # process keyboard shortcuts
        nonlocal n_keydowns
        if _n_keydowns <= n_keydowns:
            shortcut = None
        else:
            help_open = dashboard.process_shortcuts(
                display, help_open, shortcut
            )
        n_keydowns = _n_keydowns

        # based on most recent selection, fix any invalid button values
        state_module._fix_invalid_inputs(display, shortcut, dashboard.spec)

        # update url based on radio button values
        url = urls.create_new_url(
            display, defaults=dashboard.spec['default_state']
        )

        # Update visibility of interval, time, and yscale radio buttons
        raw_classes = inputs_module.compute_input_classes(
            display, dashboard.spec
        )
        classes = list(raw_classes.values())

        return list(display.values()) + classes + [url, initial_load, help_open]

    # add chart callback
    @app.callback(*update_chart_args)  # type: ignore
    def update_chart(*args) -> go.Figure:
        kwargs = dict(zip(inputs.keys(), args))
        kwargs['now'] = kwargs['now'][:10]
        state = state_module.build_state(
            _fix=False, _ui_spec=dashboard.spec, **kwargs
        )
        dashboard.create_chart(state)

    # add clientside callbacks
    inputs_module._prevent_button_focus(app)

    return app


def _compute_decorator_args(
    ui_spec: ui_types.UiSpec,
) -> tuple[list[Input | Output | State], list[Input | Output | State]]:
    dash_input_values = []
    dash_output_values = []
    dash_output_classes = []
    for name, input_spec in ui_spec['inputs'].items():
        if input_spec['type'] == 'button':
            dash_input_values.append(Input(name + '-radio', 'value'))
            dash_output_values.append(Output(name + '-radio', 'value'))
            dash_output_classes.append(Output(name + '-radio', 'className'))
        elif input_spec['type'] == 'date':
            dash_input_values.append(Input('date-picker', 'date'))
            dash_output_values.append(Output('date-picker', 'date'))
            dash_output_classes.append(Output('date-picker', 'className'))
        else:
            raise Exception('invalid type')

    update_ui_args = [
        dash_output_values,
        dash_output_classes,
        [
            Output('url', 'search'),
            Output('initial-load', 'data'),
            Output('help-modal', 'is_open'),
        ],
        dash_input_values,
        [
            Input('url', 'search'),
            State('initial-load', 'data'),
            Input('keyboard', 'n_keydowns'),
            State('keyboard', 'keydown'),
            State('help-modal', 'is_open'),
        ],
    ]

    update_chart_args = [Output('main-chart', 'figure'), dash_input_values]

    return update_ui_args, update_chart_args



from __future__ import annotations

import typing

import polars as pl
import toolstr
import plotly.graph_objects as go  # type: ignore

from .. import styles


def create_treemap(
    df: pl.DataFrame,
    title: str,
    metric: str,
    grouping: str,
    prefix: str | None = None,
    colors: typing.Mapping[str, str] = styles.colors,
) -> go.Figure:
    toolstr_kwargs: dict[str, typing.Any] = {
        'order_of_magnitude': True,
        'decimals': 1,
    }
    if prefix is not None:
        toolstr_kwargs['prefix'] = prefix

    grand_total = df['value'].sum()

    names = []
    sizes = []
    parents: list[str | None] = []
    custom_data = []

    names.append(title)
    sizes.append(grand_total)
    parents.append(None)
    custom_data.append(
        '<b>{name}</b><br>'.format(name=title)
        + toolstr.format(grand_total, **toolstr_kwargs)
        + ' total '
        + metric
    )
    top_parent = title

    if grouping == 'token':
        other_grouping = 'network'
    elif grouping == 'network':
        other_grouping = 'token'
    else:
        raise Exception('invalid grouping: ' + str(grouping))

    grouped = (
        df.group_by(grouping)
        .agg(pl.sum('value'))
        .sort('value', descending=True)
    )
    for group, value in grouped.rows():
        names.append(group)
        sizes.append(value)
        parents.append(top_parent)
        custom_data.append(
            '<b>{name}</b><br>'.format(name=group)
            + toolstr.format(value, **toolstr_kwargs)
            + ' '
            + metric
            + '<br>'
            + toolstr.format(value / grand_total, percentage=True, decimals=1)
            + ' of total'
        )

    for group, value in grouped.rows():
        other_grouped = df.filter(pl.col(grouping) == group)
        token_total = other_grouped['value'].sum()
        other_grouped = other_grouped.with_columns(
            prop=(pl.col.value / pl.col.value.sum()),
        )  # .filter(pl.col.prop > 0.001)
        if len(other_grouped) >= 2:
            for other_group, value in other_grouped[
                [other_grouping, 'value']
            ].rows():
                while other_group in names:
                    other_group = ' ' + other_group + ' '
                names.append(other_group)
                sizes.append(value)
                parents.append(group)
                if grand_total == 0:
                    grand_fraction = 0
                else:
                    grand_fraction = value / grand_total
                if token_total == 0:
                    token_fraction = 0
                else:
                    token_fraction = value / token_total
                custom_data.append(
                    (
                        '<b>{name}</b><br>'.format(
                            name=other_group + ' ' + group
                        )
                        + toolstr.format(value, **toolstr_kwargs)
                        + ' '
                        + metric
                        + '<br>'
                        + toolstr.format(
                            grand_fraction, percentage=True, decimals=1
                        )
                        + ' of total'
                        + '<br>'
                        + toolstr.format(
                            token_fraction, percentage=True, decimals=1
                        )
                        + '% of {group}'.format(group=group)
                    )
                )

    fig = go.Figure(
        go.Treemap(
            labels=names,
            parents=parents,
            values=sizes,
            textfont=dict(
                family='monospace',
                size=28,
            ),
            branchvalues='total',
            customdata=custom_data,
        ),
    )

    fig.update_traces(
        textposition='middle center',
        hovertemplate='%{customdata}<extra></extra>',
        hoverlabel={'font': {'size': 22, 'family': 'Monospace'}},
    )

    fig.update_traces(selector=dict(label=title), textfont=dict(color='red'))

    margin = 1
    fig.update_layout(
        margin=dict(t=margin, l=margin, r=margin, b=margin),
        treemapcolorway=[
            colors.get(group) for group in grouped[grouping] if group in colors
        ],
    )

    return fig



from __future__ import annotations

from typing import Any

from dash import html  # type: ignore
import dash_extensions  # type: ignore
import dash_bootstrap_components as dbc  # type: ignore
import tooltime

from . import timestamps
from . import ui_types


def _create_shortcuts_listeners(
    shortcuts: dict[str, ui_types.ShortcutSpec],
) -> dash_extensions.Keyboard:
    return dash_extensions.Keyboard(
        captureKeys=list(shortcuts.keys()), id='keyboard', n_keydowns=1
    )


def _create_help_modal(
    shortcuts: dict[str, ui_types.ShortcutSpec],
) -> dbc.Modal:
    rows = []
    for key, shortcut in shortcuts.items():
        field = shortcut['field']
        if field is not None:
            field = field.replace('_', ' ')
        # get description
        if 'help' in shortcut:
            description = shortcut['help']
        elif shortcut['action'] == 'show_help':
            description = 'toggle help'
        elif shortcut['action'] == 'select':
            description = html.Div(
                ['set ', field, ' to ', html.B(shortcut['value'])]
            )
        elif shortcut['action'] == 'cycle_next':
            description = html.Div(['go to next ', html.B(field)])
        elif shortcut['action'] == 'cycle_previous':
            description = html.Div(['go to previous ', html.B(field)])
        else:
            raise Exception('invalid shortcut action')

        # get key string
        if key == 'ArrowRight':
            key_str = '→'
        elif key == 'ArrowLeft':
            key_str = '←'
        else:
            key_str = key

        # build table row
        row = [
            html.Td(' '),
            html.Td(html.B(key_str)),
            html.Td(' '),
            html.Td(' '),
            html.Td(' '),
            html.Td(description),
            html.Td(' '),
        ]
        rows.append(row)

    return dbc.Modal(
        [
            dbc.ModalHeader(dbc.ModalTitle('Keyboard Shortcuts')),
            dbc.ModalBody(
                html.Table(html.Tbody([html.Tr(row) for row in rows])),
            ),
        ],
        id='help-modal',
        size='sm',
        is_open=False,
    )


def _process_keyboard_shortcuts(
    display: dict[str, Any],
    help_open: bool,
    raw_shortcut: dict[str, Any],
    data_date_range: tuple[tooltime.Timestamp, tooltime.Timestamp],
    ui_spec: ui_types.UiSpec,
) -> bool:
    shortcut = ui_spec['shortcuts'][raw_shortcut['key']]
    field = shortcut['field']
    if shortcut['action'] == 'select' and field is not None:
        display[field] = shortcut['value']
    elif shortcut['action'] == 'cycle_next' and field is not None:
        options = ui_spec['inputs'][field]['button_options']
        index = options.index(display[field])
        if index + 1 == len(options):
            display[field] = options[0]
        else:
            display[field] = options[index + 1]
    elif shortcut['action'] == 'cycle_previous' and field is not None:
        options = ui_spec['inputs'][field]['button_options']
        index = options.index(display[field])
        if index == 0:
            display[field] = options[-1]
        else:
            display[field] = options[index - 1]
    elif (
        shortcut['action'] in ['increment', 'decrement']
        and shortcut['field'] == 'now'
    ):
        if raw_shortcut['ctrlKey'] or raw_shortcut['shiftKey']:
            value = 'medium'
        elif raw_shortcut['ctrlKey'] and raw_shortcut['shiftKey']:
            value = 'small'
        else:
            value = 'large'

        duration = timestamps.time_increments[display['time_window']][value]
        new_now_dt = tooltime.timestamp_to_datetime(display['now'][:10])
        if shortcut['action'] == 'increment':
            new_now_dt = timestamps._increment_datetime(new_now_dt, duration)
        elif shortcut['action'] == 'decrement':
            new_now_dt = timestamps._decrement_datetime(new_now_dt, duration)
        else:
            raise Exception()
        new_now = tooltime.timestamp_to_date(new_now_dt)

        # check that new time is valid
        min_now = tooltime.timestamp_to_date(data_date_range[0])
        if new_now < min_now:
            new_now = min_now
        max_now = tooltime.timestamp_to_date(data_date_range[1])
        if new_now > max_now:
            new_now = max_now

        display['now'] = new_now
    elif shortcut['action'] == 'show_help':
        help_open = not help_open
    else:
        raise Exception('invalid shortcut action')

    return help_open



from __future__ import annotations

import datetime
import typing

import polars as pl
import tooltime

from . import ui_types


ui_interval_to_df_interval = {
    'date': 'day',
    'week': 'week',
    'month': 'month',
}


time_increments = {
    'all': {
        'large': '365d',
        'medium': '30d',
        'small': '7d',
    },
    '365d': {
        'large': '365d',
        'medium': '30d',
        'small': '7d',
    },
    '30d': {
        'large': '30d',
        'medium': '7d',
        'small': '1d',
    },
    '7d': {
        'large': '7d',
        'medium': '1d',
        'small': '1d',
    },
}


def _get_date_range(df: pl.DataFrame) -> tuple[int, int]:
    return (
        tooltime.timestamp_to_seconds(
            typing.cast(datetime.datetime, df['timestamp'].min())
        ),
        tooltime.timestamp_to_seconds(
            typing.cast(datetime.datetime, df['timestamp'].max())
        ),
    )


def _increment_datetime(
    dt: datetime.datetime, duration: str
) -> datetime.datetime:
    if duration == '1d':
        return dt + datetime.timedelta(days=1)
    elif duration == '7d':
        return dt + datetime.timedelta(days=7)
    elif duration == '30d':
        if dt.month == 12:
            return dt.replace(month=1, year=dt.year + 1)
        else:
            return dt.replace(month=dt.month + 1)
    elif duration == '365d':
        return dt.replace(year=dt.year + 1)
    else:
        raise Exception('invalid increment')


def _decrement_datetime(
    dt: datetime.datetime, duration: str
) -> datetime.datetime:
    if duration == '1d':
        return dt - datetime.timedelta(days=1)
    elif duration == '7d':
        return dt - datetime.timedelta(days=7)
    elif duration == '30d':
        if dt.month == 1:
            return dt.replace(month=12, year=dt.year - 1)
        else:
            return dt.replace(month=dt.month - 1)
    elif duration == '365d':
        return dt.replace(year=dt.year - 1)
    else:
        raise Exception('invalid increment')


def _ensure_time_window_exceeds_sample_interval(
    values: dict[str, str], trigger: str, shortcut: ui_types.ShortcutSpec
) -> None:
    if values['time_window'] == '7d' and values['sample_interval'] == 'weekly':
        if trigger == 'time_window-radio':
            values['sample_interval'] = 'daily'
        elif trigger == 'sample_interval-radio':
            values['time_window'] = '30d'
        elif trigger == 'keyboard' and shortcut['field'] == 'sample_interval':
            values['time_window'] = '30d'
        else:
            values['sample_interval'] = 'daily'
    elif (
        values['time_window'] == '7d' and values['sample_interval'] == 'monthly'
    ):
        if trigger == 'time_window-radio':
            values['sample_interval'] = 'daily'
        elif trigger == 'sample_interval-radio':
            values['time_window'] = '365d'
        elif trigger == 'keyboard' and shortcut['field'] == 'sample_interval':
            values['time_window'] = '365d'
        else:
            values['sample_interval'] = 'daily'
    elif (
        values['time_window'] == '30d'
        and values['sample_interval'] == 'monthly'
    ):
        if trigger == 'time_window-radio':
            values['sample_interval'] = 'weekly'
        elif trigger == 'sample_interval-radio':
            values['time_window'] = '365d'
        elif trigger == 'keyboard' and shortcut['field'] == 'sample_interval':
            values['time_window'] = '365d'
        else:
            values['sample_interval'] = 'weekly'



from __future__ import annotations

import typing

import tooltime
import polars as pl
import plotly.graph_objects as go  # type: ignore

from ... import plotly
from .. import state as state_module
from .. import ui_types
from .. import timestamps


def create_tree_fig(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    ui_spec: ui_types.UiSpec,
    data_date_range: tuple[int, int],
) -> go.Figure:
    df = _prepare_treemap_data(df, state, ui_spec=ui_spec)
    styles = _get_treemap_styles(df, state, ui_spec, data_date_range)
    return plotly.create_treemap(df=df, **styles)


def _prepare_treemap_data(
    df: pl.DataFrame, state: dict[str, typing.Any], ui_spec: ui_types.UiSpec
) -> pl.DataFrame:
    # filter by metric
    metric = state_module.get_df_metric(state, ui_spec)
    df = df.filter(pl.col.metric == metric)

    if metric in ui_spec['metric_types']['point_in_time']:
        now = tooltime.timestamp_to_seconds(state['now']) * 1e6
        df = df.filter(pl.col.timestamp == now)

    elif metric in ui_spec['metric_types']['unique']:
        # filter sample_interval
        ui_interval = state['sample_interval']
        interval = timestamps.ui_interval_to_df_interval[ui_interval]
        df = df.filter(pl.col.interval == interval)

        # select the closest point to now
        time_values = df['timestamp'].sort().unique()
        now = tooltime.timestamp_to_seconds(state['now']) * 1e6
        time_index = time_values.cast(pl.Int64).search_sorted(now)
        if time_index == len(time_values):
            time_index = -1
        time = time_values[time_index]
        df = df.filter(pl.col.timestamp == time)

    else:
        df = df.filter(pl.col.interval == 'day')

        # filter time window
        end_time = tooltime.timestamp_to_seconds(state['now'])
        if state['time_window'] == 'all':
            df = df.filter(pl.col.timestamp <= end_time * 1e6)
        else:
            window = tooltime.timelength_to_seconds(state['time_window'])
            start_time = end_time - window
            df = df.filter(
                pl.col.timestamp > start_time * 1e6,
                pl.col.timestamp <= end_time * 1e6,
            )

        # aggregate
        df = (
            df.filter(pl.col.metric == metric)
            .group_by('token', 'network')
            .agg(
                pl.sum('value'),
                min_timestamp=pl.min('timestamp'),
                max_timestamp=pl.max('timestamp'),
            )
        )

    return df


def _get_treemap_styles(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    ui_spec: ui_types.UiSpec,
    data_date_range: tuple[int, int],
) -> dict[str, typing.Any]:
    # create title
    title = ui_spec['helpers']['create_title'](
        df, state, data_date_range=data_date_range, ui_spec=ui_spec
    )

    # get prefix
    metric = state_module.get_df_metric(state, ui_spec=ui_spec)
    if ui_spec['metric_units'].get(metric) == '$':
        prefix = '$'
    else:
        prefix = None

    return {
        'title': title,
        'grouping': state['grouping'],
        'metric': metric,
        'prefix': prefix,
    }



from __future__ import annotations

from typing import Any

from dash import Dash, dcc, html  # type: ignore
import plotly.graph_objects as go  # type: ignore
import polars as pl

from . import apps
from . import figures
from . import timestamps
from . import shortcuts
from . import ui_types


class Dashboard:
    spec: ui_types.UiSpec
    dfs: dict[str, pl.DataFrame]
    date_ranges: dict[str, tuple[int, int]]
    debug: bool
    app: Dash
    n_keydowns: int

    def __init__(
        self,
        *,
        name: str,
        dfs: dict[str, pl.DataFrame],
        spec: ui_types.UiSpec,
        debug: bool = False,
        assets_folder: str | None = None,
    ):
        self.spec = spec
        self.dfs = dfs
        self.date_ranges = {
            name: timestamps._get_date_range(df) for name, df in dfs.items()
        }
        self.app = apps._create_app(
            name=name,
            dashboard=self,
            debug=debug,
            assets_folder=assets_folder,
        )
        self.n_keydowns = 0

    def run(
        self, port: str = '8052', jupyter_mode: str = 'external', **kwargs: Any
    ) -> None:
        self.app.run(jupyter_mode=jupyter_mode, port=port, **kwargs)

    def get_layout(self, inputs: dict[str, html.Div]) -> list[html.Div]:
        return [
            shortcuts._create_shortcuts_listeners(self.spec['shortcuts']),
            dcc.Location(id='url', refresh=False),
            dcc.Store(id='initial-load', data=True),
            html.Div(list(inputs.values()), className='radio-group-row'),
            dcc.Graph(id='main-chart', config={'responsive': True}),
            shortcuts._create_help_modal(self.spec['shortcuts']),
            html.Div(id='prevent-focus-trigger'),
        ]

    def get_dataset(self, state: dict[str, Any]) -> str:
        if len(self.dfs) == 1:
            return next(iter(self.dfs.keys()))
        else:
            raise NotImplementedError('get_dataset()')

    def process_shortcuts(
        self,
        state: dict[str, Any],
        help_open: bool,
        raw_shortcut: dict[str, Any],
    ) -> bool:
        return shortcuts._process_keyboard_shortcuts(
            display=state,
            help_open=help_open,
            raw_shortcut=raw_shortcut,
            data_date_range=self.date_ranges[self.get_dataset(state)],
            ui_spec=self.spec,
        )

    def create_chart(self, state: dict[str, Any]) -> go.Figure:
        data_name = self.get_dataset(state)
        df = self.dfs[data_name]
        date_range = self.date_ranges[data_name]

        if state['format'] in ['line', 'line %', 'area', 'area %']:
            return figures.create_time_series_fig(
                df, state, self.debug, ui_spec=self.spec
            )
        elif state['format'] == 'tree':
            return figures.create_tree_fig(
                df, state, ui_spec=self.spec, data_date_range=date_range
            )
        else:
            raise Exception('invalid format: ' + str(state['format']))



from __future__ import annotations

import typing

from dash import callback_context  # type: ignore
import tooltime

from . import timestamps
from . import ui_types


def get_df_metric(state: dict[str, str], ui_spec: ui_types.UiSpec) -> str:
    if state['metric'] in ui_spec['submetrics']:
        return state['submetric']
    else:
        return state['metric']


def build_state(
    *,
    _ui_spec: ui_types.UiSpec,
    _defaults: bool = False,
    _fix: bool = True,
    _validate: bool = True,
    **kwargs: typing.Any,
) -> dict[str, str]:
    _replace_aliased_values(kwargs, ui_spec=_ui_spec)
    state = dict(**kwargs)

    if _defaults:
        for key, value in _ui_spec['default_state'].items():
            if key not in state:
                state[key] = value
        if 'now' not in state:
            state['now'] = tooltime.now()

    if _fix or _validate:
        if _fix:
            _fix_invalid_option_combinations(state, _ui_spec)
        elif _validate:
            invalid_combos = _find_invalid_option_combinations(state, _ui_spec)
            if invalid_combos is not None:
                raise Exception(
                    'invalid combination of state: ' + str(invalid_combos)
                )
        invalid_values = _find_invalid_option_values(state, _ui_spec)
        if len(invalid_values) > 0:
            raise Exception('invalid option values: ' + str(invalid_values))

    return state


def _replace_aliased_values(
    combo: dict[str, str],
    ui_spec: ui_types.UiSpec,
) -> None:
    for field, aliases in ui_spec['aliases'].items():
        if field in combo:
            combo[field] = aliases.get(combo[field], combo[field])


def _fix_invalid_option_combinations(
    state: dict[str, str],
    ui_spec: ui_types.UiSpec,
    freeze_fields: list[str] | None = None,
    aliased: bool = False,
) -> None:
    while True:
        invalid = _find_invalid_option_combinations(state, ui_spec)
        if invalid is None:
            return
        for key in invalid.keys():
            if freeze_fields is not None and key not in freeze_fields:
                state[key] = ui_spec['default_state'][key]


def _fix_invalid_inputs(
    state: dict[str, str],
    raw_shortcut: dict[str, typing.Any],
    ui_spec: ui_types.UiSpec,
) -> None:
    # determine origin of trigger
    ctx = callback_context
    triggered_input = ctx.triggered[0]['prop_id'].split('.')[0]

    # get shortcut spec
    if raw_shortcut is not None:
        shortcut = ui_spec['shortcuts'][raw_shortcut['key']]
    else:
        shortcut = {}

    # fix time_window too small for sample_interval too small
    timestamps._ensure_time_window_exceeds_sample_interval(
        state, triggered_input, shortcut
    )

    # check invalid option combinations
    trigger = triggered_input.split('-radio')[0]

    _fix_invalid_option_combinations(
        state, freeze_fields=[trigger], aliased=True, ui_spec=ui_spec
    )


def _find_invalid_option_combinations(
    state: dict[str, str],
    ui_spec: ui_types.UiSpec,
    aliased: bool = False,
) -> dict[str, typing.Any] | None:
    if aliased:
        state = state.copy()
        _replace_aliased_values(state, ui_spec)
    for value in ui_spec['invalid_states']:
        if all(state[k] == v for k, v in value.items()):
            return value
    return None


def _find_invalid_option_values(
    state: dict[str, str],
    ui_spec: ui_types.UiSpec,
) -> dict[str, typing.Any]:
    output = {}
    for key, value in state.items():
        if key in ui_spec['aliases']:
            valid_values = list(ui_spec['aliases'][key].values())
        elif key in ui_spec['inputs']:
            if ui_spec['inputs'][key]['type'] == 'button':
                valid_values = ui_spec['inputs'][key]['button_options']
            elif ui_spec['inputs'][key]['type'] == 'date':
                continue
            else:
                raise Exception('invalid type')
        else:
            raise Exception('invalid key: ' + str(key))
        if value not in valid_values:
            output[key] = value
    return output



from __future__ import annotations

import datetime
from typing import Any, Literal, TypedDict
from typing_extensions import NotRequired


class UiSpec(TypedDict):
    default_state: dict[str, Any]
    invalid_states: list[dict[str, str]]
    inputs: dict[str, InputSpec]
    aliases: dict[str, dict[str, str]]
    reverse_aliases: dict[str, dict[str, str]]
    metric_types: MetricTypes
    metric_units: dict[str, str]
    submetrics: dict[str, list[str]]
    shortcuts: dict[str, ShortcutSpec]
    helpers: dict[str, Any]  # function


class InputSpec(TypedDict):
    type: Literal['button', 'date']
    description: str
    default: str
    visibility: NotRequired[InputVisibility]
    button_options: NotRequired[list[str]]
    date_options: NotRequired[tuple[datetime.datetime, datetime.datetime]]


class InputVisibility(TypedDict):
    start_hidden: NotRequired[bool]
    hide_if: NotRequired[list[dict[str, Any]]]
    show_if: NotRequired[list[dict[str, Any]]]
    f: NotRequired[Any]  # function


class MetricTypes(TypedDict):
    point_in_time: list[str]
    sum: list[str]
    unique: list[str]


class ShortcutSpec(TypedDict):
    action: Literal[
        'select',
        'cycle_next',
        'cycle_previous',
        'show_help',
        'increment',
        'decrement',
    ]
    field: str | None
    value: NotRequired[str]
    help: NotRequired[str]


def create_ui_spec(
    *,
    invalid_states: list[dict[str, str]],
    inputs: dict[str, InputSpec],
    aliases: dict[str, dict[str, str]],
    metric_types: MetricTypes,
    metric_units: dict[str, str],
    submetrics: dict[str, list[str]],
    shortcuts: dict[str, ShortcutSpec],
    helpers: dict[str, Any],
) -> UiSpec:
    reverse_aliases = {
        field: {v: k for k, v in aliases[field].items()}
        for field in aliases.keys()
    }

    default_state = {
        input: input_spec['default'] for input, input_spec in inputs.items()
    }

    ui_spec: UiSpec = {
        'default_state': default_state,
        'invalid_states': invalid_states,
        'inputs': inputs,
        'aliases': aliases,
        'reverse_aliases': reverse_aliases,
        'metric_types': metric_types,
        'metric_units': metric_units,
        'submetrics': submetrics,
        'shortcuts': shortcuts,
        'helpers': helpers,
    }

    validate_ui_spec(ui_spec)

    return ui_spec


def validate_ui_spec(ui_spec: UiSpec) -> None:
    # types match typeddict definitions
    pass

    # aliases match reverse aliases
    pass

    # every metric / submetric has a type
    pass

    # every metric present is defined
    pass



from __future__ import annotations

from typing import Any

import polars as pl

from ...state import get_df_metric
from ... import ui_types


def _get_timeseries_style_kwargs(
    state: dict[str, Any],
    df: pl.DataFrame,
    data_date_range: tuple[int, int],
    ui_spec: ui_types.UiSpec,
) -> dict[str, Any]:
    # get yscale
    log_y = False
    if state['format'] == 'line' and state['yscale'] == 'log':
        log_y = True

    relative_y = False
    if state['format'] == 'line' and state['yscale'] == 'relative':
        relative_y = True

    # get ytick format
    if state['format'] in ['line %', 'area %']:
        ytick_format = '%'
    elif state['metric'] in ['total_supply', 'volume']:
        ytick_format = '$'
    else:
        ytick_format = None

    # get ylabel
    ylabel = (
        ui_spec['reverse_aliases']['sample_interval'][
            state['sample_interval']
        ].lower()
        + ' '
        + ui_spec['reverse_aliases']['metric'][state['metric']].lower()
    )
    if state['format'] in ['line %', 'area %']:
        ylabel = ylabel + ' share'

    # get title
    title = ui_spec['helpers']['create_title'](
        df=df, state=state, data_date_range=data_date_range, ui_spec=ui_spec
    )

    # get xlabel
    if state['sample_interval'] == 'month':
        xlabel = 'month'
    elif state['sample_interval'] == 'week':
        xlabel = 'week'
    elif state['sample_interval'] == 'date':
        xlabel = 'date'
    else:
        raise Exception(
            'invalid selected sample_interval: ' + str(state['sample_interval'])
        )

    x = 'timestamp'
    total = False
    hover = True

    if state['format'] == 'line':
        # toggle total entry
        if state['total'] == 'total':
            total = True
        elif state['total'] == 'no total':
            total = False
        else:
            raise Exception('unknown value for total: ' + str(state['total']))

        # toggle hover
        if state['hover'] == 'hover':
            hover = True
        elif state['hover'] == 'no hover':
            hover = False
        else:
            raise Exception('unknown value for hover: ' + str(state['hover']))

        # age-specific styles
        if state['xalign'] == 'age':
            if state['sample_interval'] == 'date':
                xlabel = 'days'
            else:
                xlabel = xlabel + 's'
            xlabel = state['grouping'] + ' age (' + xlabel + ')'
            x = 'age'

        # relative styles
        if state['ynormalize'] == 'relative':
            ytick_format = '%'

    # return options based on figure type
    df_metric = get_df_metric(state, ui_spec)
    if state['format'] in ['line', 'line %']:
        return dict(
            metric=df_metric,
            grouping=state['grouping'],
            x=x,
            log_y=log_y,
            relative_y=relative_y,
            xlabel=xlabel,
            ylabel=ylabel,
            ytick_format=ytick_format,
            title=title,
            hover=hover,
            total=total,
        )
    elif state['format'] in ['area', 'area %']:
        styles: dict[str, Any] = dict(
            metric=state['metric'],
            grouping=state['grouping'],
            x=x,
            xlabel=xlabel,
            ylabel=ylabel,
            ytick_format=ytick_format,
            title=title,
        )
        if state['format'] == 'area %':
            styles['ylim'] = [0, 1]
        return styles
    else:
        raise Exception('invalid format')
